#!/usr/bin/env bash

# Install some more dependencies.
pip install python-cpl
pip install skycalc_ipy
pip install git+https://github.com/astronomyk/ScopeSim.git
pip install git+https://github.com/astronomyk/ScopeSim_Templates.git
npm install -g pajv

# Put down some default configuration.
mkdir -p ~/.awe
{
    echo "[global]"
    echo "database_engine: filebased"
    echo "database_user: AWANONYMOUS"
    echo "database_password: anonymous"
    echo "data_server=localhost"
    echo "data_port=8008"
} > ~/.awe/Environment.cfg

# Cache scopesim downloads
{
    echo "alias: SIM"
    echo "properties:"
    echo "  file:"
    echo "    use_cached_downloads: True"
} > ~/.scopesim_rc.yaml

# Setup git prompt
curl https://raw.githubusercontent.com/git/git/master/contrib/completion/git-prompt.sh -o ~/git-prompt.sh
{
    echo "source ~/git-prompt.sh"
    echo "export PS1='[\u@\h \W\$(__git_ps1 \" (%s)\")]\$ '"
    echo "conda activate base"
} >> ~/.bashrc

# Clone the project into /opt/project, because this is also what PyCharm uses.
# To be mounted over by the project directory on the host.
git clone https://gitlab.astro-wise.org/micado/micadowise.git /opt/project
conda develop /opt/project/src
ln -s /opt/project/src/micado "${CONDA_SP_DIR}"

# Make scripts executable
chmod +x /root/bin/*.sh

# Download some files to the project directory
cd /opt/project || exit

if [ ! -f ditaa-0.11.0-standalone.jar ]; then
    wget https://github.com/stathissideris/ditaa/releases/download/v0.11.0/ditaa-0.11.0-standalone.jar
fi

if [ ! -f vodslparser-0.4.0.jar ]; then
    wget https://repo1.maven.org/maven2/org/javastro/vodsl/vodslparser/0.4.0/vodslparser-0.4.0.jar
fi
