# -*- coding: utf-8 -*-
"""mcd_img_stack recipe."""

import astropy.units as u
import numpy

from micado.dataitems.Stack import Stack
from micado.dataitems.StackExtension import StackExtension


def make(self: Stack):
    """mcd_img_stack recipe."""
    # Propagate attributes.
    detrend0 = self.detrends[0]
    self.filtname = detrend0.filtname
    self.mjdobs = detrend0.mjdobs
    self.tpl__start = detrend0.tpl__start
    self.ra = detrend0.ra
    self.dec = detrend0.dec

    # Use this image as reference.
    image00 = detrend0.extensions[0].image
    image00med = float(numpy.median(image00))

    # Calculate pixel values.
    # TODO: Ensure this works with multiple Detectors.
    for detrend in self.detrends:
        assert (
            len(detrend.extensions) == 1
        ), "Stacking is only supported for single-detector images."

    # TODO: Put pixel scale in the extension header! Use CDELT etc!
    pixel_scale = 4 * u.mas / u.pix
    x_size_detrend, y_size_detrend = [a * u.pix for a in image00.shape]

    # TODO: Put RA/Dec in the extensions!
    ras = [detrend.ra * u.deg for detrend in self.detrends]
    decs = [detrend.dec * u.deg for detrend in self.detrends]
    ra_min = min(ras)
    ra_max = max(ras) + x_size_detrend * pixel_scale
    dec_min = min(decs)
    dec_max = max(decs) + y_size_detrend * pixel_scale

    ra_delta = ra_max - ra_min
    dec_delta = dec_max - dec_min
    x_delta = (ra_delta / pixel_scale).decompose()
    y_delta = (dec_delta / pixel_scale).decompose()
    shape = (int(round(x_delta.value)), int(round(y_delta.value)))

    data = numpy.full_like(
        detrend0.extensions[0].image,
        shape=shape,
        fill_value=image00med,
    )

    nod = len(self.detrends)
    for detrend in self.detrends:
        for extension in detrend.extensions:
            data_to_add = (extension.image - image00med) / nod
            ra_start = detrend.ra * u.deg - ra_min
            dec_start = detrend.dec * u.deg - dec_min
            x_start = int(round((ra_start / pixel_scale).decompose().value))
            y_start = int(round((dec_start / pixel_scale).decompose().value))
            x_size, y_size = extension.image.shape
            data[x_start : x_start + x_size, y_start : y_start + y_size] += data_to_add

    extension = StackExtension(
        NAXIS=detrend0.NAXIS,
        NAXIS1=int(round(x_delta.value)),
        NAXIS2=int(round(y_delta.value)),
        BITPIX=detrend0.BITPIX,
    )
    extension.image = data

    self.extensions = [extension]
