# -*- coding: utf-8 -*-
"""Generate an association matrix."""

import itertools
import random
from collections import OrderedDict
from dataclasses import dataclass
from functools import lru_cache
from typing import Dict, Any, List, Optional

from micado.dataitems.AstromRef import AstromRef
from micado.dataitems.Astromfield import Astromfield
from micado.dataitems.Bkgfield import Bkgfield
from micado.dataitems.CalibHCI import CalibHCI
from micado.dataitems.CalibratedCatalog import CalibratedCatalog
from micado.dataitems.CoroBackground import CoroBackground
from micado.dataitems.CoroOffset import CoroOffset

# from micado.dataitems.CoroPupilmask import CoroPupilmask
from micado.dataitems.CubeHCI import CubeHCI
from micado.dataitems.Dark import Dark
from micado.dataitems.DepersistedAstromfield import DepersistedAstromfield
from micado.dataitems.DepersistedBkgfield import DepersistedBkgfield
from micado.dataitems.DepersistedCoroBackground import DepersistedCoroBackground
from micado.dataitems.DepersistedCoroOffset import DepersistedCoroOffset
from micado.dataitems.DepersistedCoroPupil import DepersistedCoroPupil

# from micado.dataitems.DepersistedCoroPupilmask import DepersistedCoroPupilmask
from micado.dataitems.DepersistedCoroWaffle import DepersistedCoroWaffle
from micado.dataitems.DepersistedDark import DepersistedDark
from micado.dataitems.DepersistedFlat import DepersistedFlat

# from micado.dataitems.DepersistedIllum import DepersistedIllum
from micado.dataitems.DepersistedScience import DepersistedScience
from micado.dataitems.DepersistedScienceAi import DepersistedScienceAi
from micado.dataitems.DepersistedStdfield import DepersistedStdfield
from micado.dataitems.DepersistedWam import DepersistedWam
from micado.dataitems.Detrend import Detrend
from micado.dataitems.DetrendAi import DetrendAi
from micado.dataitems.DetrendAstromfield import DetrendAstromfield
from micado.dataitems.DetrendBase import DetrendBase
from micado.dataitems.DetrendBkgfield import DetrendBkgfield
from micado.dataitems.DetrendCoroBackground import DetrendCoroBackground
from micado.dataitems.DetrendFinalAi import DetrendFinalAi
from micado.dataitems.DetrendFinalAstromfield import DetrendFinalAstromfield
from micado.dataitems.DetrendFinalBase import DetrendFinalBase
from micado.dataitems.DetrendFinalHCIBase import DetrendFinalHCIBase
from micado.dataitems.DetrendFinalHCIOffset import DetrendFinalHCIOffset
from micado.dataitems.DetrendFinalHCIScienceBase import DetrendFinalHCIScienceBase
from micado.dataitems.DetrendFinalPupil import DetrendFinalPupil

# from micado.dataitems.DetrendFinalPupilmask import DetrendFinalPupilmask
from micado.dataitems.DetrendFinalStdfield import DetrendFinalStdfield
from micado.dataitems.DetrendFinalWaffle import DetrendFinalWaffle
from micado.dataitems.DetrendFinalWam import DetrendFinalWam

from micado.dataitems.DetrendHCIBase import DetrendHCIBase
from micado.dataitems.DetrendHCIOffset import DetrendHCIOffset

# from micado.dataitems.DetrendIllum import DetrendIllum

# from micado.dataitems.DetrendPupilmask import DetrendPupilmask
from micado.dataitems.DetrendStdfield import DetrendStdfield
from micado.dataitems.DetrendWaffle import DetrendWaffle
from micado.dataitems.DetrendWam import DetrendWam
from micado.dataitems.DistortionBase import DistortionBase
from micado.dataitems.DistortionElt import DistortionElt
from micado.dataitems.DistortionRefBase import DistortionRefBase
from micado.dataitems.DistortionWam import DistortionWam
from micado.dataitems.DpsicsOpt import DpsicsOpt

from micado.dataitems.Flat import Flat

# from micado.dataitems.Illum import Illum
from micado.dataitems.Nonlinearity import Nonlinearity
from micado.dataitems.PhotomBase import PhotomBase
from micado.dataitems.PhotomStdfield import PhotomStdfield
from micado.dataitems.DpsicsStaroffsetOpt import DpsicsStaroffsetOpt
from micado.dataitems.DpsicsStaroffsetSci import DpsicsStaroffsetSci
from micado.dataitems.Regrid import Regrid
from micado.dataitems.Science import Science
from micado.dataitems.ScienceAi import ScienceAi
from micado.dataitems.ScienceCoroPupil import ScienceCoroPupil
from micado.dataitems.ScienceCoroWaffle import ScienceCoroWaffle
from micado.dataitems.StarcenterBase import StarcenterBase
from micado.dataitems.StarcenterPupil import StarcenterPupil
from micado.dataitems.StarcenterWaffle import StarcenterWaffle
from micado.dataitems.Staroffset import Staroffset
from micado.dataitems.DepersistedBase import DepersistedBase
from micado.dataitems.Stdfield import Stdfield
from micado.dataitems.Wam import Wam
from micado.dataitems.WamRef import WamRef
from micado.main.Opt import Opt
from micado.main.RecipeStepExtra import RecipeStepBase
from micado.main.Sky import Sky
from micado.main.SkySiaiBase import SkySiaiBase
from micado.recipes.ApplyAstrom import ApplyAstrom
from micado.recipes.ComputePhotomInScienceFilter import ComputePhotomInScienceFilter
from micado.recipes.CorrectGeomDistortion import CorrectGeomDistortion
from micado.recipes.CreateSourceCatalogFinal import CreateSourceCatalogFinal
from micado.recipes.DeriveDpsicsDetrend import DeriveDpsicsDetrend
from micado.toolbox.containedint import ContainedInt

from micado.dataitems.Astrom import Astrom
from micado.dataitems.Background import Background
from micado.dataitems.Calibrated import Calibrated
from micado.dataitems.DetrendFinal import DetrendFinal
from micado.dataitems.IllumCorr import IllumCorr
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.Stack import Stack
from micado.main.DataItem import DataItem
from micado.toolbox.ditaa_from_wise import (
    get_ditaaclass,
    PATH_DOCS_GENERATED,
    text_from_template,
    text_from_dataitem,
    get_applicable_templates_from_dataitems,
    get_raws_derived,
)
from micado.toolbox.ditaacreator import (
    ProcessMini,
    PosArt,
    RecipeShort,
    Canvas,
    Line,
    Process,
    Template as TemplateDitaa,
    DataExternal,
    DataStatic,
)

# noinspection PyUnresolvedReferences
from micado.dataitems.DetrendPupil import DetrendPupil  # noqa

CLASSES_STANDARD_IMAGING_SMALL = [
    DetrendFinalBase,
    Astrom,
    PhotomStdfield,
    Calibrated,
    Stack,
]

CLASSES_STANDARD_IMAGING = [
    # We decided to have a one-to-one correspondence between the association
    # matrices and the pipelines. The recipes to derive Persistence and
    # Nonlinearity are therefore not in the diagram by default.
    # Persistence,
    DepersistedBase,
    MasterDark,
    MasterFlat,
    # Nonlinearity,
    DetrendBase,
    Background,
    # IllumCorr,
    # DetrendFinalBase,
] + CLASSES_STANDARD_IMAGING_SMALL

CLASSES_STANDARD_IMAGING_FINAL = [
    Calibrated,
    Stack,
]

CLASSES_STANDARD_IMAGING_ALL = CLASSES_STANDARD_IMAGING + [
    Detrend,
    # DetrendIllum,
    DetrendStdfield,
    DetrendBkgfield,
    DetrendFinal,
    DetrendFinalStdfield,
    DepersistedDark,
    DepersistedFlat,
    # DepersistedIllum,
    DepersistedScience,
    DepersistedStdfield,
    DepersistedBkgfield,
    Dark,
    Flat,
    # Illum,
    Science,
    Stdfield,
    Bkgfield,
]

CLASSES_ASTROMETRIC_IMAGING_SMALL = [
    DetrendFinalBase,
    DistortionBase,
    # DistortionWam,
    # DistortionElt,
    Astrom,
    PhotomStdfield,
    Calibrated,
    CalibratedCatalog,
    Regrid,
]

CLASSES_ASTROMETRIC_IMAGING = [
    DepersistedBase,
    MasterDark,
    MasterFlat,
    DetrendBase,
    Background,
    # IllumCorr,
    # DetrendFinalBase,
] + CLASSES_ASTROMETRIC_IMAGING_SMALL

CLASSES_ASTROMETRIC_IMAGING_FINAL = [
    CalibratedCatalog,
    Regrid,
]

CLASSES_ASTROMETRIC_IMAGING_ALL = CLASSES_ASTROMETRIC_IMAGING + [
    DistortionWam,
    DistortionElt,
    DetrendAi,
    # DetrendIllum,
    DetrendStdfield,
    DetrendBkgfield,
    DetrendWam,
    DetrendAstromfield,
    DetrendFinalAi,
    DetrendFinalStdfield,
    DetrendFinalWam,
    DetrendFinalAstromfield,
    DepersistedDark,
    DepersistedFlat,
    # DepersistedIllum,
    DepersistedScienceAi,
    DepersistedStdfield,
    DepersistedBkgfield,
    DepersistedWam,
    DepersistedAstromfield,
    Dark,
    Flat,
    # Illum,
    ScienceAi,
    Stdfield,
    Bkgfield,
    Wam,
    Astromfield,
    WamRef,
    AstromRef,
]

CLASSES_HIGH_CONTRAST_IMAGING_FOCAL = [
    # Persistence,
    DepersistedBase,
    MasterDark,
    MasterFlat,
    DetrendHCIBase,
    Background,
    DetrendFinalBase,
    Staroffset,
    StarcenterWaffle,
    # StarcenterPupil,
    CalibHCI,
    CubeHCI,
]

CLASSES_HIGH_CONTRAST_IMAGING_PUPIL = [
    # Persistence,
    DepersistedBase,
    MasterDark,
    MasterFlat,
    DetrendHCIBase,
    Background,
    DetrendFinalBase,
    Staroffset,
    # StarcenterWaffle,
    StarcenterPupil,
    CalibHCI,
    CubeHCI,
]

CLASSES_HIGH_CONTRAST_IMAGING_SMALL = [
    DetrendFinalBase,
    Staroffset,
    StarcenterWaffle,
    StarcenterPupil,
    CalibHCI,
    CubeHCI,
]

CLASSES_HIGH_CONTRAST_IMAGING = [
    # Persistence,
    DepersistedBase,
    MasterDark,
    MasterFlat,
    DetrendHCIBase,
    Background,
    # DetrendFinalBase,
] + CLASSES_HIGH_CONTRAST_IMAGING_SMALL

# TODO: Generate CLASSES_HIGH_CONTRAST_IMAGING automatically
assert set(
    CLASSES_HIGH_CONTRAST_IMAGING_FOCAL + CLASSES_HIGH_CONTRAST_IMAGING_PUPIL
) == set(CLASSES_HIGH_CONTRAST_IMAGING)

CLASSES_HIGH_CONTRAST_IMAGING_FOCAL_ALL = CLASSES_HIGH_CONTRAST_IMAGING_FOCAL + [
    Dark,
    Flat,
    # ScienceCoroPupil,
    ScienceCoroWaffle,
    CoroOffset,
    CoroBackground,
    DepersistedDark,
    DepersistedFlat,
    DepersistedCoroBackground,
    DepersistedCoroWaffle,
    # DepersistedCoroPupil,
    DepersistedCoroOffset,
    DetrendWaffle,
    # DetrendPupil,
    DetrendHCIOffset,
    DetrendFinalHCIOffset,
    DetrendCoroBackground,
    DetrendFinalWaffle,
    # DetrendFinalPupil,
]
CLASSES_HIGH_CONTRAST_IMAGING_FOCAL_ALL.remove(DetrendHCIBase)

CLASSES_HIGH_CONTRAST_IMAGING_PUPIL_ALL = CLASSES_HIGH_CONTRAST_IMAGING_PUPIL + [
    Dark,
    Flat,
    ScienceCoroPupil,
    # CoroPupilmask,
    # ScienceCoroWaffle,
    CoroOffset,
    DepersistedDark,
    DepersistedFlat,
    # DepersistedCoroWaffle,
    DepersistedCoroPupil,
    # DepersistedCoroPupilmask,
    DepersistedCoroOffset,
    # DetrendWaffle,
    DetrendPupil,
    # DetrendPupilmask,
    DetrendHCIOffset,
    DetrendFinalHCIOffset,
    # DetrendFinalWaffle,
    DetrendFinalPupil,
    # DetrendFinalPupilmask,
]
CLASSES_HIGH_CONTRAST_IMAGING_PUPIL_ALL.remove(DetrendHCIBase)

CLASSES_HIGH_CONTRAST_IMAGING_ALL = CLASSES_HIGH_CONTRAST_IMAGING + [
    Dark,
    Flat,
    ScienceCoroPupil,
    # CoroPupilmask,
    ScienceCoroWaffle,
    CoroOffset,
    CoroBackground,
    DepersistedDark,
    DepersistedFlat,
    DepersistedCoroWaffle,
    DepersistedCoroPupil,
    # DepersistedCoroPupilmask,
    DepersistedCoroOffset,
    DepersistedCoroBackground,
    DetrendCoroBackground,
    DetrendWaffle,
    DetrendPupil,
    # DetrendPupilmask,
    DetrendHCIOffset,
    DetrendFinalHCIOffset,
    DetrendFinalWaffle,
    DetrendFinalPupil,
    # DetrendFinalPupilmask,
]
CLASSES_HIGH_CONTRAST_IMAGING_ALL.remove(DetrendHCIBase)

# TODO: Generate CLASSES_HIGH_CONTRAST_IMAGING_ALL automatically
assert set(
    CLASSES_HIGH_CONTRAST_IMAGING_FOCAL_ALL + CLASSES_HIGH_CONTRAST_IMAGING_PUPIL_ALL
) == set(CLASSES_HIGH_CONTRAST_IMAGING_ALL)

CLASSES_QACITS = [
    DpsicsStaroffsetOpt,
    DpsicsStaroffsetSci,
]

CLASSES_QACITS_ALL = CLASSES_QACITS + [
    DpsicsOpt,
    # DpsicsSci,
    ScienceCoroWaffle,
]

CLASSES_UPTO_REDUCED = [
    DepersistedBase,
    MasterDark,
    MasterFlat,
    DetrendBase,
    Background,
    DetrendFinalBase,
]

CLASSES_UPTO_REDUCED_ALL = (
    CLASSES_STANDARD_IMAGING_ALL
    + CLASSES_ASTROMETRIC_IMAGING_ALL
    + CLASSES_HIGH_CONTRAST_IMAGING_ALL
)

CLASSES_UPTO_REDUCED_FINAL = []

DASHED_HORIZONTAL = False


@dataclass
class NeighbourArt:
    """Art that is a neighbour of other Art."""

    art: "ConnPosArt" = None
    distance: int = None
    line: Optional[str] = "|"


@dataclass
class ConnPosArt(PosArt):
    """Connected Position Art."""

    centered: bool = True
    up: NeighbourArt = None
    down: NeighbourArt = None
    left: NeighbourArt = None
    right: NeighbourArt = None

    @property
    def tension(self):
        """The tension felt by this PosArt."""
        tension = 0
        if self.leftp < 0:
            tension += 1000
        if self.top < 0:
            tension += 1000
        if self.top > 0:
            tension += 0.01
        if self.up is not None:
            # Prefer to have up two above us.
            if self.up.distance is not None:
                tension += (
                    2
                    / (1 + self.top / 100)
                    * abs(self.top - self.up.art.bottom - self.up.distance)
                )
            # And centered
            tension += 0.5 * abs(self.center_horizontal - self.up.art.center_horizontal)
        if self.left is not None:
            # Prefer to have left two to the left
            if self.left.distance is not None:
                tension += (
                    2
                    / (1 + self.leftp / 100)
                    * abs(self.leftp - self.left.art.rightp - self.left.distance)
                )
            # And centered
            tension += 3.5 * abs(self.center_vertical - self.left.art.center_vertical)
        if self.down is not None:
            if self.down.distance is not None:
                tension += (
                    1
                    / (1 + self.top / 100)
                    * abs(self.bottom - self.down.art.top + self.down.distance)
                )
            # And centered
            tension += abs(self.center_horizontal - self.down.art.center_horizontal)
        if self.right is not None:
            # Prefer to have left two to the left
            if self.right.distance is not None:
                tension += (
                    1
                    / (1 + self.leftp / 100)
                    * abs(self.rightp - self.right.art.leftp + self.right.distance)
                )
            # And centered
            tension += 2.5 * abs(self.center_vertical - self.right.art.center_vertical)

        return tension

    @property
    def tension_left(self):
        """Cumulative tension to the left.

        Currently unused.
        """
        rr = self
        ll = self.left
        ww = 0
        tens = 0
        while ll is not None:
            tens += abs((rr.leftp - ll.art.rightp) - (ll.distance + ww))
            ww += ll.distance + rr.art.width
            rr = ll.art
            ll = ll.art.left
        return tens


@dataclass
class ConnCanvas(Canvas):
    """Canvas with connected PosArts."""

    @property
    def tension(self):
        """The tension felt by the arts."""
        tension = sum(po.tension for po in self.arts if isinstance(po, ConnPosArt))

        for a1, a2 in itertools.permutations(self.arts, 2):
            # https://stackoverflow.com/questions/27152904/calculate-overlapped-area-between-two-rectangles
            dx = min(a1.rightp, a2.rightp) - max(a1.x, a2.x)
            dy = min(a1.bottom, a2.bottom) - max(a1.y, a2.y)
            if dx >= 0 and dy >= 0:
                tension += 4 * dx * dy

        return tension

    def shuffle(self):
        """Shuffle the Canvas.

        This algorithm does not work with ContainedInts, because the
        backing up and restoring of the position only works with normal ints.
        TODO: Ensure this algorithm works with ContainedInts.
        """

        def move_l(pob):
            """Move left."""
            pob.x -= 1

        def move_r(pob):
            """Move right."""
            pob.x += 1

        def move_u(pob):
            """Move up."""
            pob.y -= 1

        def move_d(pob):
            """Move down."""
            pob.y += 1

        moves = [move_l, move_r, move_u, move_d]
        arts_shuffled = random.sample(self.arts, len(self.arts))
        too = self.tension
        po_same, move_same = None, None
        for po in arts_shuffled:
            xo, yo, to = po.x, po.y, self.tension
            for move in random.sample(moves, len(moves)):
                move(po)
                if self.tension < to:
                    break
                if self.tension == to:
                    po_same, move_same = po, move
                    po.x, po.y = xo, yo
                else:
                    po.x, po.y = xo, yo
            if self.tension < too:
                break
        else:
            # No move led to a lower tension
            if move_same is not None:
                move_same(po_same)


def shuffle(cc: ConnCanvas, contained_ints_dict):  # sourcery skip
    """Shuffle ConnCanvas cc by permutating contained ints."""
    contained_ints = [cf for cf in contained_ints_dict.values()]
    cfs_shuffled = random.sample(contained_ints, len(contained_ints))
    too = cc.tension
    move_same = None
    for cf in cfs_shuffled:
        to = cc.tension
        moves = [object.__getattribute__(cf, mm) for mm in ["__iadd__", "__isub__"]]
        for move in random.sample(moves, len(moves)):
            move(1)
            if cc.tension < to:
                break
            if cc.tension == to:
                move_same = move
                move(-1)
            else:
                move(-1)
        if cc.tension < too:
            break
    else:
        # No move led to a lower tension
        if move_same is not None:
            move_same(1)


@lru_cache(maxsize=1000000)
def can_zig_zag(l1: str, l2: str):
    """Can we zigzag through these two lines?

    Args:
        l1: line 1
        l2: line 2

    True:
      '   xx   '
      'xx     x'
    False:
      '   xx   '
      'xxx    x'
    """
    # first step is free
    lsplus = [" " + li + " " for li in [l1, l2]]
    scores = [
        [
            cm.isspace() * (not cl.isspace())
            + 2 * (not cm.isspace())
            + cm.isspace() * (not cr.isspace())
            for cl, cm, cr in zip(li[:-2], li[1:-1], li[2:])
        ]
        for li in lsplus
    ]
    sumscores = [sum(li) for li in zip(*scores)]
    goods = [li < 3 for li in sumscores]

    return all(goods)


def remove_line(lines: List[str], contained_ints_dict):
    """If there are two consecutive empty lines in the canvas, remove one."""
    for i, (l1, l2, l3) in enumerate(zip(lines[2:], lines[1:-1], lines[:-2])):
        if l2.isspace() and can_zig_zag(l1, l3):
            break
    else:
        return False

    ypos_to_remove = i + 1
    cis_to_move: List[ContainedInt] = [
        ci
        for k, ci in contained_ints_dict.items()
        if "ypos" in k and ci > ypos_to_remove
    ]
    for ci in cis_to_move:  # NOSONAR
        ci -= 1  # NOSONAR
    return lines[: i + 1] + lines[i + 2 :]


def get_lines_up_left(cc: ConnCanvas):
    """Get lines up and left."""
    lines = []
    po: ConnPosArt
    for po in cc.arts:
        if po.up is not None and po.up.line:
            lines.append(
                ConnPosArt(
                    x=po.x - 1,
                    y=po.up.art.bottom - 1,
                    centered=False,
                    art=Line(
                        vertical=True,
                        distance=po.top - po.up.art.bottom + 2,
                        style=po.up.line,
                        connector="3",
                    ),
                )
            )
        if po.left is not None and po.left.line:
            lines.append(
                ConnPosArt(
                    x=po.left.art.rightp - 1,
                    y=po.y - 1,
                    centered=False,
                    art=Line(
                        vertical=False,
                        distance=po.leftp - po.left.art.rightp + 2,
                        style=po.left.line,
                        connector="4",
                    ),
                )
            )
    return lines


def get_duo_lines(thearts, rsteps_duoinput, big=False):
    """Get duo lines."""
    lines = []
    for rs in rsteps_duoinput:
        input_main = rs.get_other_inputdata_full()[0]
        # TODO: Refactor this into a .name_association() or so,
        #       maybe name_main()? or name_recipe()?
        if issubclass(input_main, DepersistedBase):
            name_input = DepersistedBase.name_eso()
        # TODO: Fix HCI hack!
        # elif (
        #     input_main == DetrendBase
        #     and DetrendBase.name_eso() not in thearts
        #     and DetrendHCI.name_eso() in thearts
        # ):
        #     name_input = DetrendHCI.name_eso()
        # else:
        #     name_input = rs.get_other_inputdata_full()[0].name_eso()
        do_input = rs.get_other_inputdata_full()[0]
        # TODO: Fix hack for ApplyAstrom.
        if rs == ApplyAstrom:
            do_input = DetrendFinalBase
        name_input = do_input.name_eso()
        # TODO: Fix this hack for ApplyMasterDark.
        if name_input == "DEPERSISTEDOPT_IMG":
            name_input = "DEPERSISTED_IMG"
        name_rs = rs.name_eso()
        # There are quite some exceptions to this rule, so only add the
        # duo lines if the first input is in thearts.
        if name_input not in thearts:
            from micado.recipes.ApplyMasterDark import ApplyMasterDark

            assert rs != ApplyMasterDark
            continue
        assert name_rs in thearts
        for art in thearts[name_rs]:
            # Create horizontal lines.
            lines.append(
                ConnPosArt(
                    x=thearts[name_input].rightp - 1,
                    y=thearts[name_input].y - 1,
                    centered=False,
                    art=Line(
                        vertical=False,
                        distance=art.x - thearts[name_input].rightp + 1,
                        connector="5",
                        style=":" if DASHED_HORIZONTAL else "|",
                    ),
                )
            )
            # Vertical lines.
            lines.append(
                ConnPosArt(
                    x=art.x - 1,
                    y=thearts[name_input].y - 1,
                    centered=False,
                    art=Line(
                        vertical=True,
                        distance=art.y - thearts[name_input].y + (-1 if big else 1),
                        connector="6",
                    ),
                )
            )
    # TODO: This function does nothing when big==False except make 1 dashed
    #  line.
    return lines


def is_point_in_arts(thearts, x, y):
    """Is the point at x, y within the arts?"""
    # Some arts are actually lists of arts, so flatten.
    arts2 = [a if isinstance(a, list) else [a] for a in thearts.values()]
    arts3 = [a for aa in arts2 for a in aa]
    is_ins = [a.is_inside(x, y) for a in arts3]
    return any(is_ins)


def add_lines(cc: ConnCanvas, thearts, rsteps_duoinput, big=False):
    """Add lines to connect everything."""
    lines = get_lines_up_left(cc) + get_duo_lines(thearts, rsteps_duoinput, big=big)

    for cpo in lines:
        assert isinstance(cpo.art, Line)
    line_ends = [
        ConnPosArt(
            x=cpo.x,
            y=cpo.y,
            centered=cpo.centered,
            art=Line(
                vertical=cpo.art.vertical,
                distance=0,
                connector="1",
            ),
        )
        for cpo in lines
        if not is_point_in_arts(
            thearts=thearts,
            x=cpo.x,
            y=cpo.y,
        )
    ] + [
        ConnPosArt(
            # x, y reused below
            x=cpo.x + (cpo.art.distance - 1 if not cpo.art.vertical else 0),
            y=cpo.y + (cpo.art.distance - 1 if cpo.art.vertical else 0),
            centered=cpo.centered,
            art=Line(
                vertical=cpo.art.vertical,
                distance=0,
                connector="2",
            ),
        )
        for cpo in lines
        if not is_point_in_arts(
            thearts=thearts,
            # x, y reused above
            x=cpo.x + (cpo.art.distance - 1 if not cpo.art.vertical else 0),
            y=cpo.y + (cpo.art.distance - 1 if cpo.art.vertical else 0),
        )
    ]
    for arti in cc.arts:
        assert isinstance(arti, ConnPosArt)
    arts_cc = [arti for arti in cc.arts if isinstance(arti, ConnPosArt)]

    return ConnCanvas(arts=lines + arts_cc + line_ends)


def remove_rs_text(cc: ConnCanvas):
    """Remove the text of ProcessMini circles."""
    arts_new = [
        ConnPosArt(
            x=cpo.x,
            y=cpo.y,
            centered=cpo.centered,
            art=ProcessMini(text="") if isinstance(cpo.art, ProcessMini) else cpo.art,
        )
        for cpo in cc.arts
    ]
    return ConnCanvas(arts=arts_new)


def get_duosteps(classes_to_draw):
    """Get those recipesteps that require two inputs."""
    recipesteps = set()
    for do in classes_to_draw:
        recipestep = do.recipe.prop_type.process.prop_type
        while recipestep is not None:
            recipesteps.add(recipestep)
            recipestep = recipestep.get_previous_step()
    rsteps_multiinput = [
        r for r in recipesteps if len(r.get_other_inputdata_full()) > 1
    ]
    # 4 used to be CorrectGeomDistortion from Astrom, now it is 2 as well
    # Their extra input is from a different imaging pipeline, so it does not
    # have to be in the association matrix.
    # ComputePhotomInScienceFilter has two PHOTOM_TAB's as input.
    # TODO: Fix HCI online, that is, with DpsicsStaroffsetOpt
    if DpsicsStaroffsetOpt not in classes_to_draw:
        assert {len(r.get_other_inputdata_full()) for r in rsteps_multiinput} <= {2}
    rsteps_duoinput = [
        r
        for r in recipesteps
        if len(r.get_other_inputdata_full()) == 2
        and r not in {ComputePhotomInScienceFilter}
    ]
    assert {r.get_previous_step() for r in rsteps_duoinput} <= {
        None,
        DeriveDpsicsDetrend,
        # TODO: Does having CreateSourceCatalogFinal here make sense?
        CreateSourceCatalogFinal,
    }
    assert {r.get_other_inputdata_full()[0].name_eso() for r in rsteps_duoinput} <= {
        # SI, HCI
        "DEPERSISTEDOPT_IMG",
        "DETREND_IMG",
        "DISTORTIONWAM_HDR",
        # SI
        "DETRENDFINAL_IMG",
        # AI
        "DETRENDFINALAI_IMG",
        "DETRENDFINALSIAI_IMG",
        # QACITS
        "MASTERDARK_IMG",
        "QACITSDETRENDSKP_IMG",
    }

    return rsteps_duoinput


def get_name_input_other(input_other, arts, allowed=None):
    """Get the name we should use for input_other."""
    # TODO: Fix Peristent hack
    if "depersisted" in input_other.name_eso().lower():
        return text_from_dataitem(DepersistedBase)
    elif "detrendillum" in input_other.name_eso().lower():
        return text_from_dataitem(DetrendBase)
    elif (
        "ref_photom" in input_other.name_eso().lower()
        and input_other.name_eso() in arts
    ):
        # TODO: Fix PHOTOMREF hack, e.g. check vertical dist. of dependency.
        return input_other.name_eso() + "2"
    else:
        return text_from_dataitem(input_other, allowed=allowed)


def get_bases(cc):
    """Get base classes recursively."""
    if issubclass(cc, DataItem) and "dataitems" in cc.__module__:
        yield cc
        for cci in cc.__bases__:
            yield from get_bases(cci)


def get_left_of_step(
    step,
    y_step,
    step_old,
    name_ycf_step,
    contained_ints_dict,
    xcf_left,
    arts,
    big,
    classes_final,
    classes_all,
):
    """Get what is left of this step."""
    input_others = step.get_other_inputdata_full(
        never_include_previous=not big, include_raw=False
    )
    if step_old is not None:
        left = NeighbourArt(art=step_old)
    elif step in {
        CorrectGeomDistortion,
        ComputePhotomInScienceFilter,
        DeriveDpsicsDetrend,
    }:
        # This only applies to Astrom:
        # > Astrom.recipe.prop_type.process.prop_type.previous.prop_type\
        #     .previous.prop_type.name_eso()
        # 'correct_geom_distortion'
        # TODO: This also happens in HCI and this is not a sustainable solution.
        len_max = max(len(io.name_eso()) for io in input_others)
        # noinspection PyArgumentList
        posarts2 = [
            PosArt(
                x=0,
                y=i * 4,
                # TODO: Refactor all of this so it is not necessary to overrule
                #       the ditaaclass
                # art=get_ditaaclass(input_other, final=classes_final)(
                art=DataExternal(text=input_other.name_eso().ljust(len_max)),
            )
            for i, input_other in enumerate(input_others)
        ]
        ci2 = Canvas(arts=posarts2)
        # TODO: combine with the below
        assert name_ycf_step not in contained_ints_dict
        # Move down half the height of the input
        y_step += len(input_others) * 2
        contained_ints_dict[name_ycf_step] = ContainedInt(y_step)
        cpo_oi = ConnPosArt(
            x=xcf_left if big else contained_ints_dict["xpos_0"],
            y=contained_ints_dict[name_ycf_step],
            centered=True,
            art=ci2,
        )
        arts[input_others[-1].name_eso()] = cpo_oi
        left = NeighbourArt(art=cpo_oi)
        # Move down other half the height of the input
        y_step += len(input_others) * 2
    elif input_others:
        # TODO: Fix DetrendFinal
        # assert len(input_others) == 1
        input_other = input_others[-1]
        name_input_other = get_name_input_other(
            input_other=input_other, arts=arts, allowed=classes_all
        )
        # Check whether the input is already in the graph.
        # Also check derived classes, especially if input is a Base clase,
        # e.g. Calbrated.photom is a PhotomBase, but preferably refer
        # to PhotomStdfield in the matrix because that makes it more complete.
        # TODO: However, this means that PHOTOMSTDFIELD ends up in the graph
        # even though it might be better to have PHOTOM there.
        derived_input_classes1 = input_other.get_derived_classes()
        derived_input_classes = [
            cls for cls in derived_input_classes1 if cls in classes_all
        ]

        # Some sanity checks to ensure such cases can only be introduced
        # deliberately. Expand checks when necessary.
        assert derived_input_classes or input_other != PhotomBase
        assert not derived_input_classes or input_other in [
            PhotomBase,
            DetrendFinalBase,
            StarcenterBase,
            DetrendBkgfield,
            DetrendFinalHCIBase,
            DistortionRefBase,
            DetrendFinalHCIScienceBase,
        ]
        # Also check base classes, e.g. DetrendWaffle might not be in the graph
        # but DetrendHCI (its superclass) might.
        base_input_classes = list(get_bases(input_other))
        names_to_check = [name_input_other] + [
            get_name_input_other(c, arts)
            for c in derived_input_classes + base_input_classes
        ]
        # TODO: This is too similar to the logic in get_connposart_dataitem(),
        #       where raw_existing_names is created, so these need to be
        #       refactored.
        names_ok_old = [name for name in names_to_check if name in arts]

        if names_ok_old:
            # E.g. prevent finding MASTERDARK_IMGDARK_RAW which is just
            # a repeat at the top of the big matrices.
            names_ok = names_ok_old
        else:
            derived_input_classes_to_check = derived_input_classes or [input_other]
            raw_labels = [r.name_eso() for r in derived_input_classes_to_check]
            names_ok1 = [
                name
                for name, posart in arts.items()
                if isinstance(posart, ConnPosArt)
                and set(raw_labels) - set(posart.art.text.split()) == set()
            ]
            names_ok = [
                # TODO: Fix this hack to prevent finding MASTERDARK_IMGDARK_RAW
                name
                for name in names_ok1
                if name.count("_") < 2
            ]

        assert len(set(names_ok)) <= 1 or input_other in [
            StarcenterBase,
        ]
        if len(set(names_ok)) > 1 and names_ok[0] != name_input_other:
            # There are multiple to chose from, but not the name specified
            # so refuse to chose at all.
            names_ok = []

        oi: ConnPosArt = arts.get(names_ok[0]) if names_ok else None

        if oi is not None:
            left = NeighbourArt(art=oi, distance=2)
        else:
            # Since this DataItem is not yet in the graph, it must have been
            # created elsewhere. So it is either Static, not created in any
            # pipeline. Those are hardcoded. Or it is created in another
            # pipeline, and get_ditaaclass has no way of knowing.
            clsn_oi = get_ditaaclass(input_other, final=classes_final)
            if clsn_oi != DataStatic:
                clsn_oi = DataExternal
            assert name_ycf_step not in contained_ints_dict
            contained_ints_dict[name_ycf_step] = ContainedInt(y_step)
            # noinspection PyArgumentList
            cpo_oi: ConnPosArt = ConnPosArt(
                x=xcf_left if big else contained_ints_dict["xpos_0"],
                y=contained_ints_dict[name_ycf_step],
                centered=True,
                art=clsn_oi(text=name_input_other),
            )
            arts[name_input_other] = cpo_oi
            left = NeighbourArt(art=cpo_oi)
    else:
        left = None
        # y_step += 10

    return left, y_step


def add_step(
    step,
    y_step,
    arts,
    contained_ints_dict,
    cf_xposes,
    cpo_prev,
    classes_final,
    classes_all,
    big=False,
):
    """Add a processing step to the matrix."""
    cf_xpos_left, cf_xpos = cf_xposes
    if step.name_eso() in arts:
        step_old = arts[step.name_eso()][-1]
    else:
        arts[step.name_eso()] = []
        step_old = None

    name_ycf_step = "ypos_{}".format(step.name_eso())

    left, y_step = get_left_of_step(
        step=step,
        y_step=y_step,
        step_old=step_old,
        name_ycf_step=name_ycf_step,
        contained_ints_dict=contained_ints_dict,
        xcf_left=cf_xpos_left,
        arts=arts,
        classes_final=classes_final,
        classes_all=classes_all,
        big=big,
    )
    if left is None and big:
        # A big matrix with all DRL Functions is being created.
        assert name_ycf_step not in contained_ints_dict
        contained_ints_dict[name_ycf_step] = ContainedInt(y_step)
        cf_ypos_step = contained_ints_dict[name_ycf_step]
    elif left is None:
        # A small matrix without DRL Functions is created, so only
        # include those that actually have input.
        if step.name_eso() in arts and not len(arts[step.name_eso()]):
            # Remove DRL Function.
            arts.pop(step.name_eso())
        # Return old data. Note that y_step is increased.
        return None, y_step - 3, arts, contained_ints_dict, cf_xpos, cpo_prev
    else:
        assert isinstance(left.art.y, ContainedInt)
        cf_ypos_step: ContainedInt = left.art.y
        y_step = max(cf_ypos_step.actual_int, y_step)

    assert isinstance(cf_ypos_step, ContainedInt)

    # TODO: Generalize this and remove 'depersist'
    # TODO: fix ":", it causes problems with the horizontal lines, so
    #       None is used for now.
    line_style = (
        None
        if isinstance(cpo_prev.art, RecipeShort)
        and "depersist" not in cpo_prev.art.text.lower()
        and cpo_prev.art.text != "micado_det_persistence"
        and "qacits" not in step.name_eso()
        else "|"
    )
    up = NeighbourArt(
        art=cpo_prev,
        distance=2
        if step.name_eso().endswith("_layers") or (cf_ypos_step - cpo_prev.bottom < 8)
        else None,
        line=line_style,
    )
    art = Process(text=step.name_eso()) if big else ProcessMini(text=step.name_eso())
    cpo_step = ConnPosArt(
        x=cf_xpos,
        y=cf_ypos_step,
        art=art,
        centered=True,
        up=up,
        left=left,
    )
    arts[step.name_eso()].append(cpo_step)
    cpo_prev_new = cpo_step
    y_step += 9 if big else 7
    return step, y_step, arts, contained_ints_dict, cf_xpos, cpo_prev_new


def get_connposart_dataitem(
    do,
    contained_ints_dict,
    arts,
    cf_xposes,
    cpo_recipe_old,
    classes_final,
    classes_all,
    big=False,
    allowed_templates=None,
):
    """Get ConnPosArt for DataItem do."""
    assert hasattr(do, "recipe")
    assert hasattr(do, "get_raw_class")
    cf_xpos, cf_xpos_left = cf_xposes

    y_offset_steps_cheat = 0
    # TODO: The autoincrement of of ystep is broken, so all dataitems should
    #  be listed here...
    y_offset_steps_cheat_dict: Dict[Any, int] = {
        DepersistedBase: 15,
        MasterDark: 32,
        IllumCorr: 70,
        Background: 68,
        # PhotomRef: 58,
        Nonlinearity: 110,
        DetrendFinal: 120,
        DetrendFinalBase: 120,
        DistortionBase: 150,
        Astrom: 185,
        PhotomStdfield: 220,
        Calibrated: 240,
        CalibratedCatalog: 270,
        Regrid: 270,
        Stack: 270,
        # Starcenter is aligned such that the lines don't cross the boxes.
        StarcenterWaffle: 195,
        StarcenterPupil: 215,
        Staroffset: 180,
        CalibHCI: 240,
        CubeHCI: 270,
    }

    recipe = do.recipe.prop_type
    ditaa_recipe = RecipeShort(text=recipe.name_eso())

    step = recipe.process.prop_type
    steps = []
    while step:
        steps.append(step)
        step = step.get_previous()
        if not issubclass(step, RecipeStepBase):
            step = False

    raw = do.get_raw_class()
    progenitor = do.get_progenitor_class()
    # if raw.name_eso() not in arts:
    if True:
        template = raw.template.prop_type

        # Ensure that e.g. raw darks cannot be detrended.
        # raws = raw.get_derived_classes()
        raws = get_raws_derived(do)

        if do.__name__.endswith("Base"):
            # E.g. McdImgDetrendFinal could have any DetrendBase as input, but
            # it only makes sense for DetrendStdfield or Detrend.
            assert do in [
                DistortionBase,
                DepersistedBase,
                DetrendBase,
                DetrendFinalBase,
                DetrendHCIBase,
            ], do
            derived_dos = do.get_derived_classes()
            raws_derived = [di.get_progenitor_class() for di in derived_dos]
            raws_allowed = set(classes_all) & set(raws_derived)
        else:
            raws_allowed = set(classes_all)

        templates_raws = [template] + [ra.template.prop_type for ra in raws]
        allowed_templates_raw = set(templates_raws) & set(allowed_templates)

        text_template = text_from_template(template, allowed=allowed_templates_raw)
        ditaa_template = TemplateDitaa(text_template)

        # ditaa_raw = DataRaw(text=text_from_dataitem(raw))
        text_progenitor = text_from_dataitem(
            progenitor, allowed=raws_allowed, strict=True
        )
        clsn_prog = get_ditaaclass(progenitor, final=classes_final)
        ditaa_raw = clsn_prog(text=text_progenitor)

        # Try to find an existing art that corresponds to this progenitor.
        # That is, if DEPERSISTEDILLUM_IMG, DEPERSISTEDSCIENCE_IMG or
        # DEPERSISTEDSTDFIELD_IMG can be used as input, and there is an output
        # box with DEPERSISTEDDARK_IMG, DEPERSISTEDFLAT_IMG,
        # DEPERSISTEDILLUM_IMG, DEPERSISTEDSCIENCE_IMG, and
        # DEPERSISTEDSTDFIELD_IMG, then use that one.
        raw_labels = ditaa_raw.text.split()
        raw_existing_names = [
            name
            for name, posart in arts.items()
            if isinstance(posart, ConnPosArt)
            and set(raw_labels) - set(posart.art.text.split()) == set()
            # TODO: Perhaps only allow names with 1 underscare? E.g. to
            # differentiate between
            # ['DETRENDFINAL_IMG', 'ASTROM_HDRSKYSIAI_RAW']
        ]
        y = contained_ints_dict["ypos_raw" if big else "ypos_recipe"]
        if raw_existing_names and not big:
            # if raw_existing_names:
            # A sanity check. There can be more than 1 existing box that can
            # be reused, but let's make that explicit.
            # TODO: Perhaps remove this check? It seems to only be annoying
            #       and does not seem to actually help anything.
            assert len(raw_existing_names) == 1 or raw in [
                DetrendFinal,
                Opt,
                Sky,
                Science,
                CoroOffset,
                # Why is Stdfield there?
                Stdfield,
                SkySiaiBase,
            ]
            raw_existing_name = raw_existing_names[0]
            raw_existing = arts[raw_existing_name]
            y = raw_existing.y
        else:
            raw_existing = None

        cpo_raw = ConnPosArt(
            x=cf_xpos,
            y=y,
            art=ditaa_raw,
            centered=True,
            left=NeighbourArt(
                art=raw_existing, distance=2, line=":" if DASHED_HORIZONTAL else "|"
            )
            if raw_existing
            else None,
        )
        # arts[raw.name_eso()] = cpo_raw
        arts[do.name_eso() + raw.name_eso()] = cpo_raw
        if big:
            cpo_template = ConnPosArt(
                x=cf_xpos,
                y=contained_ints_dict["ypos_template"],
                art=ditaa_template,
                centered=True,
            )
            arts[f"{template.name_eso()}_{cf_xpos}"] = cpo_template
            cpo_raw.up = NeighbourArt(art=cpo_template, distance=2)
    else:
        # The raw already exists. This probably means that this Recipe is
        # not derived from raws at all.
        ditaa_raw = None
        cpo_raw = None
        ditaa_template = None
        y_offset_steps_cheat = 40

    y_offset_steps_cheat = y_offset_steps_cheat_dict.get(do, y_offset_steps_cheat)

    cpo_recipe = ConnPosArt(
        x=cf_xpos,
        y=contained_ints_dict["ypos_recipe" if big else "ypos_raw"],
        art=ditaa_recipe,
        centered=True,
        left=NeighbourArt(art=cpo_recipe_old, distance=2, line=None)
        if cpo_recipe_old is not None
        else None,
        up=NeighbourArt(art=cpo_raw, distance=2)
        if (cpo_raw is not None) and big
        else None,
    )
    if cpo_raw is not None and big:
        cpo_raw.down = NeighbourArt(art=cpo_recipe, distance=2)
    assert recipe.name_eso() not in arts
    arts[recipe.name_eso()] = cpo_recipe

    cpo_prev = cpo_recipe if big else cpo_raw
    y_step = contained_ints_dict["ypos_recipe"] + 10 + y_offset_steps_cheat
    cpos_steps = []
    for step in steps[::-1]:
        step, y_step, arts, contained_ints_dict, cf_xpos, cpo_prev = add_step(
            step,
            y_step,
            arts,
            contained_ints_dict,
            (cf_xpos_left, cf_xpos),
            cpo_prev,
            classes_final=classes_final,
            classes_all=classes_all,
            big=big,
        )
        if step is None:
            # big is False and there is no other input to this step
            continue
        cpos_steps.append(cpo_prev)
    ditaas_steps = [cpo.art for cpo in cpos_steps]

    clsn = get_ditaaclass(do, final=classes_final)
    # noinspection PyArgumentList
    ditaa_do = clsn(text=text_from_dataitem(do, allowed=classes_all, strict=True))
    name_ycf_do = "ypos_{}".format(do.name_eso())
    assert name_ycf_do not in contained_ints_dict
    contained_ints_dict[name_ycf_do] = ContainedInt(y_step + 5)
    cpo_do = ConnPosArt(
        x=cf_xpos,
        y=contained_ints_dict[name_ycf_do],
        centered=True,
        art=ditaa_do,
        up=NeighbourArt(art=cpo_prev, distance=2),
    )
    widths = (
        [ditaa_do.width, ditaa_recipe.width]
        + ([ditaa_raw.width] if ditaa_raw is not None else [])
        + ([ditaa_template.width] if ditaa_template is not None else [])
        + [d.width for d in ditaas_steps]
    )
    width_max = max(widths)
    return cpo_do, cpo_recipe, width_max


def build_association_matrix(classes_to_draw, classes_final, classes_all, big=False):
    """Create Association Matrix."""
    contained_ints_dict = {
        "ypos_template": ContainedInt(10),
        "ypos_raw": ContainedInt(30),
        "ypos_recipe": ContainedInt(50),
        "xpos_0": ContainedInt(15),
    }
    allowed_templates = get_applicable_templates_from_dataitems(classes_all)
    arts = OrderedDict()
    cpo_recipe_old = None

    xpos_do = 30
    do: DataItem
    for doi, do in enumerate(classes_to_draw):
        assert hasattr(do, "recipe")
        assert hasattr(do, "get_raw_class")
        cf_xpos = ContainedInt(xpos_do)
        contained_ints_dict[f"xpos_{doi + 1}"] = cf_xpos
        cf_xpos_left = contained_ints_dict[f"xpos_{doi}"]
        # noinspection PyTypeChecker
        cpo_do, cpo_recipe, width_max = get_connposart_dataitem(
            do=do,
            contained_ints_dict=contained_ints_dict,
            arts=arts,
            cf_xposes=(cf_xpos, cf_xpos_left),
            cpo_recipe_old=cpo_recipe_old,
            classes_final=classes_final,
            classes_all=classes_all,
            big=big,
            allowed_templates=allowed_templates,
        )
        width_left = width_max - width_max // 2
        width_right = width_max // 2 + 1
        xpos_do += width_left
        xpos_do += width_right
        # One extra in order to be able to add arrows to the figures.
        xpos_do += 1
        cf_xpos += width_left
        assert do.name_eso() not in arts
        arts[do.name_eso()] = cpo_do

        cpo_recipe_old = cpo_recipe

    artsll = [a if isinstance(a, list) else [a] for a in arts.values()]
    artsl = [a for al in artsll for a in al]
    c = ConnCanvas(arts=list(artsl))
    return c, arts, contained_ints_dict


def generate_association_matrix(classes_to_draw, classes_final, classes_all, big=False):
    """Shuffle and add legend etc."""
    c, arts, contained_ints_dict = build_association_matrix(
        classes_to_draw, classes_final, classes_all, big=big
    )
    rsteps_duoinput = get_duosteps(classes_to_draw)

    legend_big = r"""

    +---------------+
    |               |
    | xxxxxxxxx_RAW |  Raw data
    |c999           |
    +---------------+

    +---------------+
    |               |
    | xx_EXT xx_CAT |  External data
    |c6CE           |
    +---------------+

    +---------------+
    |               |
    | xxxxxxxxx_IMG |  Intermediate data / Recipe data product
    |c47A           |
    +---------------+

    +---------------+
    |               |
    | xxxxxxxxx_IMG |  Final data
    |c283           |
    +---------------+



    +--------------+
    |              |
    | mcd_xxx_xxxx |  CPL Recipe
    |cB56          |
    +--------------+

    /-----\
    |cB56 |       DRL Function
    \-----/

    /-----\
    |cB56 +-- step to the right is identical
    \-----/

    /-----\
    |cB56 | - step to the right is different
    \-----/

+--------------+
|              |
| micado_xx_xx |  CPL Recipe
|cB56          |
+--------------+



      |
    -----   data product on the left is not input to the recipe step below
      |

      |
    --+--   data product on the left is input to the recipe step below
      |

    """
    legend_small = r"""

    +---------------+
    |               |
    | xxxxxxxxx_RAW |  Raw data
    |c999           |
    +---------------+

    +---------------+
    |               |
    | xx_EXT xx_CAT |  External data
    |c6CE           |
    +---------------+

    +---------------+
    |               |
    | xxxxxxxxx_IMG |  Intermediate data / Recipe data product
    |c47A           |
    +---------------+

    +---------------+
    |               |
    | xxxxxxxxx_IMG |  Final data
    |c283           |
    +---------------+



    +--------------+
    |              |
    | mcd_xxx_xxxx |  CPL Recipe
    |cB56          |
    +--------------+



      |
    -----   data product on the left is not input to the recipe step below
      |

      |
    --+--   data product on the left is input to the recipe step below
      |

    """

    # TODO: Add these back once supported again.
    # ------ direct data dependency
    #
    # =----- indirect data dependency;
    #        actual input data is output from other recipe

    # This way it is only necessary to call .lines only once.
    lines = c.lines
    while lines := remove_line(lines, contained_ints_dict):
        # Nothing to do here, because the line is already removed.
        pass

    sss = str(remove_rs_text(add_lines(c, arts, rsteps_duoinput, big=big)))

    # Remove empty columns
    cols = zip(*sss.splitlines())
    cols_crop = [col for col in cols if any(c != " " for c in col)]
    sss_crop = "".join("".join(c) + "\n" for c in zip(*cols_crop))

    # TODO: Fix hack to get dashed lines
    if DASHED_HORIZONTAL:
        sss_crop = sss_crop.replace("|-", "|=")
    else:
        sss_crop = sss_crop.replace("|-", "+-")

    if len(sss_crop.splitlines()) < 82:
        return sss_crop

    legend = legend_big if big else legend_small
    if not big:
        legend = ""
    lineslegend = reversed(
        [
            lleg + lart[len(lleg) :] if lleg is not None else lart
            for lart, lleg in itertools.zip_longest(
                reversed(sss_crop.splitlines()),
                reversed(legend.splitlines()),
            )
        ]
    )
    return "\n".join(lineslegend)


def main():
    """Create matrix and write to file.

    Convert with

    java -jar ditaa-0.11.0-standalone.jar associationmatrix.ditaa -E \
        --no-shadows associationmatrix.png

    or

    java -jar ditaa-0.11.0-standalone.jar associationmatrix.ditaa -E \
        --no-shadows --svg associationmatrix.svg
    """
    name_classes = {
        # TODO: Fix Online HCI DPSICS or remove?
        # "qacits": (CLASSES_QACITS, [], CLASSES_QACITS_ALL),
        "hci": (CLASSES_HIGH_CONTRAST_IMAGING, [], CLASSES_HIGH_CONTRAST_IMAGING_ALL),
        "focal": (
            CLASSES_HIGH_CONTRAST_IMAGING_FOCAL,
            [],
            CLASSES_HIGH_CONTRAST_IMAGING_FOCAL_ALL,
        ),
        "pupil": (
            CLASSES_HIGH_CONTRAST_IMAGING_PUPIL,
            [],
            CLASSES_HIGH_CONTRAST_IMAGING_PUPIL_ALL,
        ),
        "si": (
            CLASSES_STANDARD_IMAGING,
            CLASSES_STANDARD_IMAGING_FINAL,
            CLASSES_STANDARD_IMAGING_ALL,
        ),
        "ai": (
            CLASSES_ASTROMETRIC_IMAGING,
            CLASSES_ASTROMETRIC_IMAGING_FINAL,
            CLASSES_ASTROMETRIC_IMAGING_ALL,
        ),
        "sismall": (
            CLASSES_STANDARD_IMAGING_SMALL,
            CLASSES_STANDARD_IMAGING_FINAL,
            CLASSES_STANDARD_IMAGING_ALL,
        ),
        "hcismall": (
            CLASSES_HIGH_CONTRAST_IMAGING_SMALL,
            [],
            CLASSES_HIGH_CONTRAST_IMAGING_ALL,
        ),
        "aismall": (
            CLASSES_ASTROMETRIC_IMAGING_SMALL,
            CLASSES_ASTROMETRIC_IMAGING_FINAL,
            CLASSES_ASTROMETRIC_IMAGING_ALL,
        ),
        "reduced": (
            CLASSES_UPTO_REDUCED,
            CLASSES_UPTO_REDUCED_FINAL,
            CLASSES_UPTO_REDUCED_ALL,
        ),
    }
    for name, (classes, classes_final, classes_all) in name_classes.items():
        for big in [
            False,
            True,
        ]:
            s_association = generate_association_matrix(
                classes, classes_final, classes_all, big=big
            )
            print(s_association)
            postfix_name = "_" + name if name != "si" else ""
            postfix_big = "_big" if big else ""
            path_associationmatrix = (
                PATH_DOCS_GENERATED
                / f"associationmatrix{postfix_name}{postfix_big}.ditaa"
            )
            with open(path_associationmatrix, "w") as f:
                f.write(s_association)


if __name__ == "__main__":
    main()
